# -*- coding: Latin-1 -*-
#
# SMM Share and Money Management
# Copyright (c) 2007-2009 Carsten Grohmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the
# License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

## \file gui/helper.py
#
# \brief Functions and classes to simplify recurring actions
#
# \namespace gui::helper
# \brief Functions to support GUI functions e.g. to format values
# 
# \author Carsten Grohmann <mail (at) carstengrohmann (dot) de>
# 
# \par License:
# GPL version 2.0
#
# \version $Id: helper.py,v 1.35 2009-04-27 13:40:05 carsten Exp $
#
# \par Revision History:
# \code
# $Log: helper.py,v $
# Revision 1.35  2009-04-27 13:40:05  carsten
# formatMoney():
# - add places parameter
# - make currency mandatory
#
# Revision 1.34  2009-04-18 19:13:43  carsten
# - add formatPosition()
#
# Revision 1.33  2009-04-14 20:18:28  carsten
# formatDuration():
# - return a better (more detailed) duration string
# formatShareKind():
# - reduce number of return statements
#
# Revision 1.32  2009-04-09 20:51:21  carsten
# - cleanup code using pylint
#
# Revision 1.31  2009-04-06 13:36:26  carsten
# - typo in license statement
#
# Revision 1.30  2009-04-06 07:56:19  carsten
# - update copyright string
# - unify python coding line
#
# Revision 1.29  2009-02-26 19:07:20  carsten
# formatDecimal() and formatMoney():
# - [FIX] format Decimal("0") too
#
# Revision 1.28  2009-02-17 19:51:48  carsten
# - add Share.kind certificate
#
# Revision 1.27  2009-01-23 20:31:27  carsten
# - convert "normal" strings into unicode strings
# \endcode

import datetime
import decimal
import re
import wx

from storm.locals import *

import smm


## Precompiled RE pattern to match two or more white spaces or
#  single linebreak characters
RE_WHITESPACES = re.compile(r'\s{2,}|[\r\n]')

## Number of days in a year
#
#  \see formatDuration()
DAYS_PER_YEAR = decimal.Decimal('365')

## A dictionary which can lookup values by key, and keys by value.
#
# All values and keys must be hashable, and unique.
#
# \par Example:
# \code
# >>> a = gui.helper.ReverseDict(((1,2),(3,4)))
# >>> print a[1]
# 2
# >>> print a.reverse[2]
# 1
# >>> a["123"] = 67
# >>> print a["123"]
# 67
# >>> print a.reverse[67]
# 123
# \endcode
#
# \par Idea:
#
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/415903
class ReverseDict(dict):

    ## \var reverse
    #  Dictionary to store the reverse data
    #
    #  The key / value pairs are reverse stored. This means the value is
    #  used as key and the origin key is stored as value in this
    #  dictionary. Thereby an fast reverse lookup to get the
    #  corresponding key of a given value.
   
    ## Initialise the data structures.
    def __init__(self, *args, **kw):
        dict.__init__(self, *args, **kw)
        self.reverse = dict((reversed(list(i)) for i in self.items()))

    ## Set a new key / value pair
    #
    #  Already existing keys or values will be removed before adding new
    #  one.
    def __setitem__(self, key, value):
        if self.has_key(key):
            old_value = self[key]
            del self.reverse[old_value]
        dict.__setitem__(self, key, value)
        self.reverse[value] = key

    ## Remove a key / value pair out of both dictionaries.
    def __delitem__(self, key):
        value = self[key]
        del self.reverse[value]
        dict.__delitem__(self, key)

## Returns additional note to a share - short formatted
#
#  \param share   The smm.objects.Share object
#  \param max_len Length of the note
#
#  \see smm.objects.Share.notes
#  \see formatShareNotesLong()
def formatShareNotesShort(share, max_len=25):
    res = u''
    if hasattr(share, 'notes') and share.notes:
        res = share.notes

        # remove linebreaks, multi spaces  and cutt off
        res = RE_WHITESPACES.sub(u' ', res)
        res = res[:max_len]

    return res

## Returns additional note to a share - long formatted
#
#  \see smm.objects.Share.notes
#  \see formatShareNotesShort()
def formatShareNotesLong(share):
    if hasattr(share, 'notes') and share.notes:
        return share.notes
    return u''

## Returns the sign of a currency or an empty string
#
#  \see smm.objects.Currency.sign
#  \see formatCurrencyShort()
#  \see formatCurrencyLong()
def formatCurrencySign(currency):
    if currency:
        return currency.sign or u''
    return u''

## Returns the currency sign or code or an empty string
#
#  \see smm.objects.Currency.sign
#  \see formatCurrencySign()
#  \see formatCurrencyLong()
def formatCurrencyShort(currency):
    if currency:
        if currency.sign:
            return currency.sign
        else:
            return currency.code
    return u''

## Returns the formatted currency or an empty string
#
#  \see smm.objects.Currency
#  \see formatCurrencySign()
#  \see formatCurrencyShort()
def formatCurrencyLong(currency):
    if currency:
        if currency.description:
            return currency.description
        else:
            return currency.code
    return u''

## Returns the formatted date or an empty string
def formatDate(date, format='%d.%m.%Y'):
    if date:
        return date.strftime(format).decode()
    return u''


## Returns the formatted decimal value or an empty string
#
#  \param value   The decimal value
#  \param places  Required number of places after the decimal point
def formatDecimal(value, places=None):
    res = u'' 
    if isinstance(value, decimal.Decimal):
        if places:
            res = moneyfmt(value, dp=',', places=3, sep='', curr='')
            res = res.decode()
        else:
            res = u'%s' % value
    return res


## Returns the given duration formatted.
#
#  \param duration Duration in days as a \c decimal.Decimal.
#  \see DAYS_PER_YEAR
def formatDuration(duration):
    res = u''
    days = duration % DAYS_PER_YEAR
    years = duration // DAYS_PER_YEAR
    if not years:
        if not days:
            res = _(u'Keine')
        elif days == 1:
            res = _(u'1 Tag')
        else:
            res = _(u'%d Tage' % days)
    elif years == 1:
        if not days:
            res = _(u'1 Jahr')
        elif days == 1:
            res = _(u'1 Jahr und 1 Tag')
        else:
            res = _(u'1 Jahr und %d Tage' % days)
    else:
        if not days:
            res = _(u'%d Jahre' % years)
        elif days == 1:
            res = _(u'%d Jahre und 1 Tag' % years)
        else:
            res = _(u'%d Jahre und %d Tage' % (years, days))
    return res


## Returns the event formatted.
#
#  \see smm.objects.DailyQuote.event
def formatEvent(event):
    res = u''
    if event:
        if event.startswith(u'split:'):
            res = _(u'Aktiensplit %s' % event.split(u':', 1)[1])
        elif event.startswith(u'bid'):
            res = _(u'Geldkurs')
        elif event.startswith(u'ask'):
            res = _(u'Briefkurs')
        else:
            res = event
    return res


## Returns the formatted exchange or an empty string
def formatExchange(exchange):
    if exchange:
        return exchange.description
    return u''


## Returns the formatted decimal value incl. the currency or an
#         empty string
#
#  \param value     The decimal value
#  \param currency  A smm.objects.Currency instance
#  \param places    Required number of places after the decimal point
def formatMoney(value, currency, places=3):
    res = u''
    if isinstance(value, decimal.Decimal):
        res = moneyfmt(value, dp=',', places=places, sep='')
        res = res.decode()
        res += u' %s' % formatCurrencyShort(currency)
    return res


## Returns the formatted decimal value incl. percent sign or an
#         empty string
#
#  \param value     The decimal value
#  \param places    Required number of places after the decimal point
def formatPercent(value, places=3):
    res = u''
    if isinstance(value, decimal.Decimal):
        res = moneyfmt(value, dp=',', places=places, sep='')
        res = res.decode()
        res += u' %'
    return res


## Returns the formatted position or an empty string
#
#  \see smm.objects.Position
def formatPosition(pos):
    if not pos:
        return u''
        
    pattern = _(u'%s (%s - %s)')
    trans = pos.transactions.order_by(Asc(smm.Transaction.date))
    date_first = formatDate(trans.first().date)
    if pos.open:
        date_last = u''
    else:
        date_last = formatDate(trans.last().date)
    return pattern % (formatShare(pos.share), date_first, date_last)


## Returns the formatted share or an empty string
#
#  \see smm.objects.Share
#  \see formatShareKind()
def formatShare(share):
    if share:
        return share.name
    return u''


## Returns the type of the share or an empty string
#
#  \see smm.objects.Share.kind
#  \see formatShare()
def formatShareKind(share):
    res = u''
    if share and share.kind != None:
        if share.kind == 'share':
            res = _('Aktie')
        elif share.kind == 'index':
            res = _('Index')
        elif share.kind == 'fond':
            res = _('Fond')
        elif share.kind == 'etf':
            res = _('ETF')
        elif share.kind == 'warrant':
            res = _('Optionsschein')
        elif share.kind == 'certificate':
            res = _('Zertifikat')
        elif share.kind == 'bond':
            res = _('Anleihe')
        else:
            res = _('Unbekannte Gattung: %s' % share.kind)
    return res


## Returns the transaction type as a string or an empty string
#
#  \see smm.objects.Transaction.kind
def formatTransactionKind(transaction):
    if transaction and transaction.kind != None:
        if transaction.kind == 'buy':
            return _('Kauf')
        elif transaction.kind == 'sell':
            return _('Verkauf')
        elif transaction.kind == 'dividend':
            return _('Dividende')
        elif transaction.kind == 'split':
            return _('Aktiensplit')
        return _('Unbekannter Typ: %s' % transaction.kind)
    return u''


## Convert Decimal to a money formatted string.
#
#  \param value    the decimal value
#  \param places   required number of places after the decimal point
#  \param curr     optional currency symbol before the sign (may be blank)
#  \param sep      optional grouping separator (comma, period, space, or blank)
#  \param dp       decimal point indicator (comma or period)
#                   only specify as blank when places is zero
#  \param pos      optional sign for positive numbers: '+', space or blank
#  \param neg      optional sign for negative numbers: '-', '(', space or blank
#  \param trailneg optional trailing minus indicator:  '-', ')', space or blank
# \par Example:
# \code
# >>> d = Decimal('-1234567.8901')
# >>> moneyfmt(d, curr='$')
# '-$1,234,567.89'
# >>> moneyfmt(d, places=0, sep='.', dp='', neg='', trailneg='-')
# '1.234.568-'
# >>> moneyfmt(d, curr='$', neg='(', trailneg=')')
# '($1,234,567.89)'
# >>> moneyfmt(Decimal(123456789), sep=' ')
# '123 456 789.00'
# >>> moneyfmt(Decimal('-0.02'), neg='<', trailneg='>')
# '<0.02>'
# \endcode
#
# \par Source:
# Receipes of the Python decimal module
def moneyfmt(value, places=2, curr='', sep=',', dp='.',
             pos='', neg='-', trailneg=''):
    q = decimal.Decimal((0, (1,), -places))    # 2 places --> '0.01'
    sign, digits, exp = value.quantize(q).as_tuple()
    assert exp == -places    
    result = []

    # convert to list
    digits = map(str, digits)

    # create two aliases
    build, next = result.append, digits.pop

    if sign:
        build(trailneg)

    # add digits to result and fill empty positions by '0'
    for i in range(places):
        if digits:
            build(next())
        else:
            build('0')

    # add decimal point
    build(dp)

    # add leading zero for 1 < value < -1 
    if not digits:
        build('0')

    # build part before decimal point
    i = 0
    while digits:
        build(next())
        i += 1
        if i == 3 and digits:
            i = 0
            build(sep)

    # add currency and sign
    build(curr)
    if sign:
        build(neg)
    else:
        build(pos)

    # compose result and strip spaces
    result.reverse()
    result = ''.join(result)
    return result.strip()


## Convert a date from \c wxDateTime to Python \c datetime.date
#
#  \see pydate2wxdate()
def wxdate2pydate(wxdate):
    return datetime.date(
        wxdate.GetYear(),
        wxdate.GetMonth()+1,
        wxdate.GetDay()
        )

## Convert a date from Python \c datetime.date to \c wxDateTime 
#
#  \see wxdate2pydate()
def pydate2wxdate(pydate):
    return wx.DateTimeFromDMY(
        day=pydate.day,
        month=pydate.month - 1,
        year=pydate.year
        )
