# -*- coding: Latin-1 -*-
#
# SMM Share and Money Management
# Copyright (c) 2007-2010 Carsten Grohmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

## \file smm/operations.py
#
# \short High level functions to do something
#
# \author Carsten Grohmann <mail (at) carstengrohmann (dot) de>
#
# \par License:
# GPL version 2.0

import logging

from storm.locals import *

from objects import *
import database
import quote


## \brief Returns the primary quotation source for a share
#
#  \return ID of the primary data source or \c None
#  \see \ref quote.source2class
def getPrimarySourceID(share):

    # check for valid and already registered source id
    if share.source and quote.source2class.has_key(share.source):
        return share.source

    # get source id depending on the share kind
    source_id = None
    if share.kind == 'fond':
        if share.kind == 'fond' and share.isin and \
           share.name.upper().startswith(u'RAIFFEISEN'):
            source_id = 300
        elif share.isin:
            source_id = 400
        elif share.wkn:
            source_id = 200
    elif share.kind == 'share':
        if share.symbol:
            source_id = 100
    elif share.kind == 'warrant':
        if share.isin or share.wkn:
            source_id = 500

    return source_id

## \brief Returns the name of the primary quotation source for a share
#
#  \see getPrimarySourceID() and quote.source2name 
def getPrimarySourceName(share):
    res = u''
    source_id = getPrimarySourceID(share)
    if source_id:
        if source_id in quote.source2name:
            res = quote.source2name[source_id]
        else:
            res = _(u'Unbekannte Quelle #%d' % source_id)
        if not share.source:
            res = _(u'Auto (%s)' % res)
    return res


########################################################################
#
# Functions to query quotations
#
########################################################################

## \brief Fetch new daily quotes
#
#  New quotes will fetch for all selected shares. The data provider is
#  selected per share. 
#
#  \param openPos    \c True to fetch quotes of all open positions
#  \param watchList  \c True to fetch quotes of all watch list entries
#  \param add_shares List additional shares to fetch too
#
#  \see quote.source2class
def fetchDailyQuotes(openPos=True, watchList=True, add_shares=None):

    # compose lists
    share_list = []
    if openPos:
        for position in Position.get(open=True):
            share_list.append(position.share)
    if watchList:
        [ share_list.append(watch.share) for watch in Watch.get()]
    if add_shares:
        share_list.extend(add_shares)

    # make list entries unique
    share_list = list(set(share_list))

    # sort shares by data provider
    provider_list = {}
    for share in share_list:
        source_id = getPrimarySourceID(share)
        if source_id != None and source_id in quote.source2class:
            source_class = quote.source2class[source_id]
            if not provider_list.has_key(source_class):
                provider_list[source_class] = []
            provider_list[source_class].append(share)
        else:
            logging.warning(
                _(u'Es wurde keine passende Quelle fr Kurse des Wertpapiers "%s" gefunden.'),
                share.name
                )

    # fetch data
    for provider in provider_list:
        try:
            quote.fetchQuote(provider, provider_list[provider])
        except:
            logging.exception(_(u'Es trat ein unerwareter Fehler auf.'))
