# -*- coding: Latin-1 -*-
#
# SMM Share and Money Management
# Copyright (c) 2007-2010 Carsten Grohmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the
# License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

## \file smm/quote/__init__.py
#
# Summarise all stuff in one point
#
# \namespace smm::quote
# Functions to import and register data tracker modules
#
# \author Carsten Grohmann <mail (at) carstengrohmann (dot) de>

# import general python modules
import logging
import os.path
import sys
import zipfile

# import project modules
from smm.objects import *

## Mapping of source id to a FetchPlugin 
#
#  \see \ref getPrimarySourceID()
#  \see source2name
source2class = {}

## Mapping of the source id to the name of the data fetcher
#  \see  source2class
source2name = {}

## Search, import and register FetchPlugins
#
#  \see _modules_from_zip()
#  \see _modules_from_filesystem()
#  \see source2class
#  \see source2name
def search_and_register():
    modules = []
    logger = logging.getLogger('smm.quote.search_and_register')

    # get a list of all modules to process
    for element in sys.path:
        if element.endswith('smm.zip'):
            modules = _modules_from_zip(element)
            break

    # no zip file
    else:
        modules = _modules_from_filesystem()

    if not modules:
        logging.info(_(u'Es wurden keine Plugins gefunden.'))
        logging.debug(
            _(u'sys.path: %s'),
            sys.path
            )
        return

    # process modules
    for module_name in modules:
        try:
            full_module_name = 'smm.quote.%s' % module_name
            logger.debug(
                _(u'Versuche das Pluging "%s" zu importieren'),
                full_module_name
                )
            module = __import__(full_module_name, globals(), {}, [module_name])
        except (ImportError, SyntaxError):
            logger.warning(
                _(u'Konnte das Module "%s" nicht importieren'),
                full_module_name,
                )
        except:
            logger.exception(
                _(u'Es trat ein unerwarteter Fehler auf. Details wurden in das Fehlerprotokoll geschrieben.')
                )

 
        # search tracker classes
        attr_list = dir(module)
        for attr_name in attr_list:
            logger.debug(
                _(u'Suche Plugin im Modul "%s"'),
                attr_name.decode('ascii', 'replace'),
                )

            klass = getattr(module, attr_name)

            # ignore all non-classes
            if not hasattr(klass, "__base__"):
                continue

            # ignore all classes not deriavated from FetchPluginBase
            if klass.__base__.__name__ != "FetchPluginBase":
                continue

            # ingore wrong interface version
            if not hasattr(klass, 'interface_version'):
                logger.debug(
                    _(u'Ignoriere altes Plugin "%s"'),
                    klass.__name__.decode('ascii', 'replace')
                    )
                continue

            if klass.interface_version != 2:
                logger.debug(
                    _(u'Ignoriere Plugin "%(name)s" da die Version der Schnittstelle %(version)d statt 2 ist.'),
                    {
                        'name':    klass.__name__.decode('ascii', 'replace'),
                        'version': klass.interface_version,
                    }
                    )
                continue

            if not hasattr(klass, 'getCurrent'):
                logger.warning(
                    _(u'Plugin "%s" fehlt die Funktion zum Abholen der Daten.'),
                    klass.__name__.decode('ascii', 'replace'),
                    )
                continue

            # class should be ok -> register
            source2class[klass.provider_id] = klass
            source2name[klass.provider_id] = klass.name

            logger.info(
                _(u'Fetcher-Plugin "%s" wurde registiert.'),
                klass.name,
                )


## Extract a list of plugins from a module zip archive
#
#  \param filename Path and Name of the module zip archive
#
#  \return List of module names
def _modules_from_zip(filename):
    _modules = []
    if not zipfile.is_zipfile(filename):
        return _modules
    zip_archive = zipfile.ZipFile(filename)
    for name in zip_archive.namelist():
        if name.startswith('smm/quote/') and name.endswith('.py') and \
           name != 'smm/quote/__init__.py':
            _modules.append(os.path.basename(name)[:-3])
    return _modules

## Search in the installation directory for plugings
#
#  \return List of module names
def _modules_from_filesystem():
    quote_dir = os.path.dirname(__file__)
    _modules = []

    if not os.path.isdir(quote_dir):
        logging.warning(
            _(u'Das Pluginverzeichnis %s existiert nicht. Es werden keine Plugins geladen.'),
            quote_dir
            )
        return _modules

    for name in os.listdir(quote_dir):
        if name.endswith('.py') and name != '__init__.py':
            _modules.append(name[:-3])

    return _modules

## Generic function to fetch quotes using a FetchPlugin
#
#  The quotations are fetched and converted to \ref smm.objects.DailyQuote
#  "daily quotes" automatically.
#
#  \param klass      Class of the FetchPlugin
#  \param share_list List of \ref smm.objects.Share "Shares"
def fetchQuote(klass, share_list):

    logger = logging.getLogger('smm.quote.fetchQuote')
    request_dict = {}
    key2share = {}

    # query quotations
    fetch = klass()
    logger.info(u'Hole Kursdaten von %s', fetch.name)

    # build request list
    for share in share_list:
        if not share.isin:
            logger.info(u'Es ist keine ISIN fr "%s" bekannt.', share.name)
            continue

        isin = share.isin

        request_dict[isin] = {}
        request = request_dict[isin]
        request['share.isin'] = isin

        currency = share.getPrefCurrency()
        if currency:
            request['currency.code'] = currency.code
        if share.wkn:
            request['share.wkn'] = share.wkn
        if share.symbol:
            request['share.symbol'] = share.symbol
        if share.exchange:
            request['exchange.symbol'] = share.exchange.symbol

        key2share[isin] = share

    response = fetch.getCurrent(request_dict)

    # store returned data into DailyQuote objects
    for key in response:
        request   = request_dict[key]
        isin      = request['share.isin']
        share     = key2share[key]
        quotation = response[key]
        pref_currency_code = request.get('currency.code', None)
        pref_currency_sign = request.get('currency.sign', None)

        if not quotation['success']:
            if quotation['errormsg']:
                logger.error(
                    _(u'Es sind keine Kursdaten fr "%s" verfgbar. Details: %s'),
                    share.name,
                    quotation['errormsg'],
                    )
            else:
                logger.error(
                    _(u'Es sind keine Kursdaten fr "%s" verfgbar.'),
                    share.name,
                    )
            continue

        # convert to Python objects
        new_data = {}
        new_data['close']  = quotation['quote.price']
        new_data['date']   = quotation['quote.datetime'].date()
        new_data['source'] = fetch.provider_id
        new_data['share']  = share

        # convert city of exchange into Exchange objects
        if 'exchange.name' in quotation:
            city = quotation.get('exchange.name')
            if city:
                ex = Exchange.get(city=city)
                if ex:
                    new_data['exchange'] = ex
                else:
                    logger.warning(
                        _(u'Die Kursdaten enthalten den unbekannten Brsenplatz "%s".'),
                        city
                        )
            else:
                logger.warning(
                    _(u'Die Kursdaten fr "%s" enthalten einen leeren Brsenplatz.'),
                    share.name,
                    )
        elif 'exchange.symbol' in quotation:
            ex = Exchange.get(symbol=quotation['exchange.symbol'])
            if ex:
                new_data['exchange'] = ex
            else:
                logger.warning(
                    _(u'Die Kursdaten enthalten den unbekannten Brsenplatz "%s".'),
                    quotation['exchange.symbol'],
                    )
        elif share.kind != u'fond':
            logger.debug(
                _(u'Die Kursdaten fr "%s" sind ohne Brsenplatz.'),
                share.name,
                )

        if 'currency.code' in quotation:
            cur = Currency.get(code=quotation['currency.code'])
            if cur:
                new_data['currency'] = cur
            else:
                logger.warning(
                    _(u'Die Kursdaten fr "%s" enthalten den ungltigen Whrungscode "%s".'),
                    share.name,
                    quotation['currency.code'].decode('ascii'),
                    )
            # check currency equility
            if pref_currency_code and quotation['currency.code'] != pref_currency_code:
                logger.warning(
                    _(u'Die Kursdaten fr "%s" haben die falsche Whrung (%s statt %s).'),
                    share.name,
                    quotation['currency.code'],
                    pref_currency_code,
                    )
        elif 'currency.sign' in quotation:
            cur = Currency.get(sign=quotation['currency.sign'])
            if cur:
                new_data['currency'] = cur
            else:
                logger.warning(
                    _(u'Die Kursdaten fr "%s" enthalten ein ungltiges Whrungszeichen "%s".'),
                    share.name,
                    quotation['currency.sign'],
                    )
            # check currency equility
            if pref_currency_sign and quotation['currency.sign'] != pref_currency_sign:
                logger.warning(
                    _(u'Die Kursdaten fr "%s" haben die falsche Whrung (%s statt %s).'),
                    share.name,
                    quotation['currency.sign'],
                    pref_currency_sign,
                    )
        else:
            logger.warning(
                _(u'Die Kursdaten fr "%s" sind ohne Whrung.'),
                share.name,
                )

        # Map various fetched properties to DailyQuote properties
        for f_prop, dq_prop in (
            ('quote.close',  'close'),
            ('quote.high',   'high'),
            ('quote.low',    'low'),
            ('quote.open',   'open'),
            ('quote.volume', 'volume'),
            ):
            if f_prop in quotation:
                new_data[dq_prop] = quotation[f_prop]

        # check minimum attribute set
        min_attrs = set([
            'close',
            'currency',
            'date',
            'source',
            ])
        cur_attrs = set(new_data)
        miss_attrs = min_attrs - cur_attrs
        if miss_attrs:
            logger.warning(
                _(u'Die Kursdaten fr "%s" sind unvollstndig. Es fehlen die Attribute %s'),
                share.name,
                u', '.join(miss_attrs),
                )
            continue

        # check for updates
        daily_quote = DailyQuote.get(share=share, date=new_data['date'])

        # create new quotation object if none exists
        if not daily_quote:
            daily_quote = DailyQuote(share=share, date=new_data['date'])
            daily_quote.add2db(False)

        # set new / update values
        for attr in new_data:
            setattr(daily_quote, attr, new_data[attr])

        # write all changes into the database
        database.store.commit()

        logger.info(
            _(u'Es wurden Kursdaten vom %s fr "%s" abgeholt.'),
            daily_quote.date,
            daily_quote.share.name,
            )
